type var = string

type binop = Add | Mul | And | Or | Leq | Eq

type exp =
  | Int of int
  | Var of var
  | Binop of binop * exp * exp
  | If of exp * exp * exp
  | Fun of string * ty * exp (* fun (x : ty) -> e *)
  | App of exp * exp         (* function application *)
  | Let of string * exp * exp

(* Types *)
and ty =
  | TyInt
  | TyBool
  | TyFun of ty * ty (* function type ty1 -> ty2 *)

type value =
  | VInt of int
  | VBool of bool
  | VClosure of env * var * exp (* function value *)

(* (Evaluation) environment : maps variables to their values *)
and env = (var * value) list

(* Type environment : maps variables to their types *)
type tyenv = (var * ty) list

(* Evalauate an expression to a value *)
let rec eval env = function
  | Int k -> VInt k
  | Var v ->
    begin
      try List.assoc v env
      with Not_found ->
        failwith @@ Printf.sprintf "Variable `%s` is not bound" v
    end
  | Binop (op, e1, e2) ->
    begin match op, eval env e1, eval env e2 with
      | Add, VInt v1, VInt v2 -> VInt (v1 + v2)
      | Mul, VInt v1, VInt v2 -> VInt (v1 * v2)
      | And, VBool v1, VBool v2 -> VBool (v1 && v2)
      | Or, VBool v1, VBool v2 -> VBool (v1 || v2)
      | Leq, VInt v1, VInt v2 -> VBool (v1 <= v2)
      | Eq, VInt v1, VInt v2 -> VBool (v1 = v2)
      | _, _, _ -> failwith "Type error!"
    end
  | If (cond, bthen, belse) ->
    begin match eval env cond with
      | VBool true -> eval env bthen
      | VBool false -> eval env belse
      | _ -> failwith "Type error!"
    end
  | Fun (var, ty, body) ->
    VClosure (env, var, body)
  | App (func, arg) ->
    begin match eval env func, eval env arg with
      | VClosure (cls_env, var, body), v -> eval ((var, v)::cls_env) body
      | _, _ -> failwith "Type error!"
    end
  | Let (var, rhs, body) ->
    eval ((var, eval env rhs)::env) body

(* Check is the judgement tyenv |- exp : typ is derivable *)
let rec typecheck tyenv exp ty = match exp with
    (*  
        ---------------- n in Int
        tyenv |- n : Int
     *)
  | Int _ -> ty = TyInt

    (*  
        ---------------- tyenv(v) = ty
        tyenv |- v : ty
     *)
  | Var v ->
    begin
      try List.assoc v tyenv = ty
      with Not_found ->
        failwith @@ Printf.sprintf "Variable `%s` is not bound" v
    end

    (*  tyenv |- e1 : int     tyenv |- e2 : int
        ---------------------------------------
               tyenv |- e1 + e2 : int


        tyenv |- e1 : int     tyenv |- e2 : int
        ---------------------------------------
               tyenv |- e1 * e2 : int
     *)
  | Binop (Add, e1, e2) | Binop (Mul, e1, e2) ->
    ty = TyInt && typecheck tyenv e1 TyInt && typecheck tyenv e2 TyInt

    (*  tyenv |- e1 : bool     tyenv |- e2 : bool
        -----------------------------------------
               tyenv |- e1 /\ e2 : bool


        tyenv |- e1 : bool     tyenv |- e2 : bool
        -----------------------------------------
               tyenv |- e1 \/ e2 : bool
     *)
  | Binop (And, e1, e2) | Binop (Or, e1, e2) ->
    ty = TyBool && typecheck tyenv e1 TyBool && typecheck tyenv e2 TyBool

    (*  tyenv |- e1 : int     tyenv |- e2 : int
        ---------------------------------------
               tyenv |- e1 <= e2 : bool


        tyenv |- e1 : int     tyenv |- e2 : int
        ---------------------------------------
               tyenv |- e1 = e2 : bool
     *)
  | Binop (Leq, e1, e2) | Binop (Eq, e1, e2) ->
    ty = TyBool && typecheck tyenv e1 TyInt && typecheck tyenv e2 TyInt

  (*
        tyenv |- cond : bool   tyenv |- bthen : t    tyenv |- belse : t
        ----------------------------------------------------------------
               tyenv |- if cond then bthen else belse : t
     *)
  | If (cond, bthen, belse) ->
     typecheck tyenv cond TyBool && typecheck tyenv bthen ty && typecheck tyenv belse ty

  (*
                 tyenv{var -> var_ty} |- body : t
        --------------------------------------------------
         tyenv |- fun (var : var_ty) -> body : var_ty -> t
     *)
  | Fun (var, var_ty, body) ->
    begin match ty with
      | TyFun (dom, cod) -> dom = var_ty && typecheck ((var,var_ty)::tyenv) body cod
      | _ -> false
    end

    (*
         tyenv |- func : ty_arg -> ty  tyenv |- arg : ty_arg
        ----------------------------------------------------
                      tyenv |- func arg : ty

        All the rules we have seen so far are /syntax-directed/ in the sense
        that the premises can be derived syntactically from the conclusion.
        Application is our first example of a rule that is /not/ syntax-directed:
        ty_arg could be any type.  In order to type check function application,
        we will /infer/ the type of the argument, and then use that type to
        type-check the function.  (It would work just as well to infer the type
        of func and then check the type of arg).
     *)
  | App (func, arg) ->
    begin match typeinfer tyenv arg with
      | Some ty_arg ->
        typecheck tyenv func (TyFun (ty_arg, ty))
      | None -> false
    end

  | Let (var, rhs, body) ->
     (*
         tyenv |- rhs : ty_rhs   tyenv{var -> ty_rhs} |- body : ty
        ----------------------------------------------------------
                   tyenv |- let var = rhs in body : ty
      *)
    begin match typeinfer tyenv rhs with
      | Some ty_rhs ->
        typecheck ((var, ty_rhs)::tyenv) body ty
      | None -> false
    end

(* Find the unique type typ such that tyenv |- exp : typ is derivable, if one exists *)
and typeinfer tyenv exp = match exp with
  | Int _ -> Some TyInt
  | Var v ->
    begin
      try Some (List.assoc v tyenv)
      with Not_found ->
        failwith @@ Printf.sprintf "Variable `%s` is not bound" v
    end
  | Binop (Add, e1, e2) | Binop (Mul, e1, e2) ->
    if typecheck tyenv e1 TyInt && typecheck tyenv e2 TyInt then
      Some TyInt
    else
      None
  | Binop (And, e1, e2) | Binop (Or, e1, e2) ->
    if typecheck tyenv e1 TyBool && typecheck tyenv e2 TyBool then
      Some TyBool
    else
      None
  | Binop (Leq, e1, e2) | Binop (Eq, e1, e2) ->
    if typecheck tyenv e1 TyInt && typecheck tyenv e2 TyInt then
      Some TyBool
    else
      None
  | If (cond, bthen, belse) ->
    begin match typeinfer tyenv bthen with
      | Some ty_then ->
        if typecheck tyenv cond TyBool && typecheck tyenv belse ty_then then
          Some ty_then
        else
          None
      | None -> None
    end
  | Fun (var, ty_var, body) ->
    begin match typeinfer ((var,ty_var)::tyenv) body with
      | Some ty_body -> Some (TyFun (ty_var, ty_body))
      | None -> None
    end
  | App (func, arg) ->
    begin match typeinfer tyenv func with
      | Some (TyFun (ty_arg, ty_ret)) ->
        if typecheck tyenv arg ty_arg then
          Some ty_ret
        else
          None
      | _ -> None
    end
  | Let (var, rhs, body) ->
    begin match typeinfer tyenv rhs with
      | Some ty_var ->
        typeinfer ((var,ty_var)::tyenv) body
      | None -> None
    end

(* fun (x : int) -> if x <= 0 then -x else x *)
let abs = Fun ("x",
               TyInt,
               If (Binop (Leq, Var "x", Int 0),
                   Binop (Mul, Int (-1), Var "x"),
                   Var "x"))
let compose =
  Fun ("f",
       TyFun (TyInt, TyInt),
       (Fun ("g",
             TyFun (TyInt, TyInt),
             (Fun ("x",
                   TyInt,
                   (Let ("gx", App (Var "g", Var "x"), App (Var "f", Var "gx"))))))))

let curry_add =
  Fun ("x",
       TyInt,
       Fun ("y",
            TyInt,
            Binop (Add, Var "x", Var "y")))

let foo = App (App (App (compose, App (curry_add, Int 1)), App (curry_add, Int 10)), Int 0)

